/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : main.c
 Author     : Jose Luis Soncco-Alvarez and Mauricio Ayala-Rincon
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>//time
#include <mpi.h>
#include <math.h>
#include <string.h>
#include "operadores.h"
#include "ordenar_pob.h"
#include "calc_fitness.h"

#define WORKTAG 1
#define DIETAG 2
#define EVALTAG 3
#define GOTAG 4

#define DEBUG 0
#define TAM_BUFFER 1000

void leer_permutacion(permutacion *perm);
void leer_permutacion_de_archivo(char* nombre_archivo, permutacion *perm);
void mostrar_poblacion(poblacion *pob, permutacion *perm, int nro_generacion);


void leer_permutacion(permutacion *perm){
	int i,n;
	scanf("%d",&n);//leer tamanho de permutacion
	perm->tamanho = n;
	for(i=0; i<n; i++){
		scanf("%d",&perm->pi[i]);
	}
}

void leer_permutacion_de_archivo(char* nombre_archivo, permutacion *perm){
    int i,n;
    FILE *filePointer;
    if ((filePointer = fopen(nombre_archivo,"r")) == NULL){
        printf("Input file could not be opened\n");
    }
    else{
        fscanf(filePointer,"%d",&n);//leer tamanho de permutacion
        perm->tamanho = n;
        for(i=0; i<n; i++){
            fscanf(filePointer,"%d",&perm->pi[i]);
        }
    }
    fclose(filePointer);
}

void mostrar_poblacion(poblacion *pob, permutacion *perm, int nro_generacion){
	
    int i,j;
    printf("*** Generacion %d ***\n", nro_generacion);
    //imprimir poblacion actual
    for(i=pob->basePoblacion; i<pob->topePoblacion; i++){
        printf("{");
        for(j=0; j<perm->tamanho-1; j++){
            printf("%d, ", pob->cromosomas[i].genes[j]);
        }
        printf("%d", pob->cromosomas[i].genes[perm->tamanho-1]);
        printf("} fit: %d\n",pob->cromosomas[i].fitness);
    }
    //imprimir descendencia
    printf("-------------------\n");
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
        printf("{");
        for(j=0; j<perm->tamanho-1; j++){
            printf("%d, ", pob->cromosomas[i].genes[j]);
        }
        printf("%d", pob->cromosomas[i].genes[perm->tamanho-1]);
        printf("} fit: %d\n",pob->cromosomas[i].fitness);
    }
	
}



//flag em uma roda vai comecar no inicio outro no final
int Imigracao( int imigracao, int sizePopulation, int fitSemelhante, poblacion *pob, int flag){
    int i,indice=0;
    int porcento=sizePopulation *0.2;
    switch (imigracao)
    {
        case 1: // pior
            indice= (sizePopulation - porcento) +(rand() % porcento);
            break;
        case 2: //aleatorio
            indice=(rand() % sizePopulation);
            break;
        case 3://semelhante
            indice=(rand() % sizePopulation);
            if (flag)
            {
                for (i=0; i<sizePopulation; i++)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end if
            else
            {
                for (i=(sizePopulation-1); i>=0; i--)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end else
            break;
    }//end switch
    
    return indice;
    
    
}// end function Imigracao


//busca os indicies a ser enviado para outra ilha
int * buscaIndex(int qtdIndividual, int typeIndividual, int sizePopulation)
{
    int i,j;
    int * indices= malloc (qtdIndividual * sizeof (int));
    if (indices==NULL)
        printf("ATENCAO PERCA DE MEMORIA, GA_Paralle_pop.c, LINE 236\n");
    switch (typeIndividual) {
        case 1://melhor
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=i;
            }
            break;
            
        case 2: // pior
            i=0;
            j=(sizePopulation-1);
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=j;
                j--;
                
            }//end for
            break;
            
        case 3: //aleatorio
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=(rand() % sizePopulation);
            }
            break;
    }//end switch
    
    
    return indices;
    
}// end buscaIndex




void sendALL( MPI_Comm communicator, int world_rank, poblacion *pob)
{
    
    int world_size, j, iii,i;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    
    // If we are the root process, send our best data to everyone
    enviar.tamanho=pob->cromosomas[0].tamanho;
    enviar.fitness=pob->cromosomas[0].fitness;
    for (i=0; i<pob->cromosomas[0].tamanho; i++)
        enviar.pi[i]=pob->cromosomas[0].genes[i];
    
    
    
    if (world_rank!=0)
        
    {
        int error_code= MPI_Send(&enviar,//the data
                                 sizeof(permutacion),
                                 MPI_BYTE,
                                 0,//destination process
                                 0,//tag
                                 communicator);
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
    }
    
    else
    {
        for( iii = 0; iii < world_size; iii++)
        {
            
            
            if (iii != world_rank)
            {
                int error_code=MPI_Recv(&recebe,
                                        sizeof(permutacion),
                                        MPI_BYTE,
                                        iii, // processo que enviou
                                        0,//tag
                                        communicator,
                                        MPI_STATUS_IGNORE);
                
                if (error_code != MPI_SUCCESS)
                {
                    char error_string[BUFSIZ];
                    int length_of_error_string;
                    MPI_Error_string(error_code, error_string, &length_of_error_string);
                    printf("%3d: %s\n", world_rank, error_string);
                }//end if
                
                int controle=0;
                for (j=0; j<pob->tamanhoPoblacion; j++)
                {
                    if (recebe.fitness < pob->cromosomas[j].fitness)
                    {
                        pob->cromosomas[j].fitness=recebe.fitness;
                        for (i=0; i<pob->cromosomas[j].tamanho; i++)
                        {
                            pob->cromosomas[j].genes[i]=recebe.pi[i];
                        }//end for
                        controle=1;
                    }//end if
                    if (controle)
                        break;
                }//end for
            }//end if
        }//end for
        
    }//else
    
}// end function

//return 1 to insert and 0 to discard
// Pgene  vai de 0.1 ate 1
// uma flag sera imposta a fim de configura a quantidade da populacao a ser comparada
int CompareIndividuals(poblacion *pob, permutacion genomaAlvo, float Pgenes) {
    int i,j;
    int flag=1;
    int contador=0;
    int intervalo=pob->tamanhoPoblacion * Pgenes;
    //printf("intervalo levado em consideracao %d sobre %d\n ", intervalo, pob->sizePopulation);
    for (i=0; i<intervalo; i++) {
        
        for (j=0; j<pob->cromosomas[i].tamanho; j++)
        {
            if(pob->cromosomas[i].genes[j]==genomaAlvo.pi[j])
            {    contador++;
            }
        }//end for
        
        if (contador==pob->cromosomas[i].tamanho) {
            flag=0;
            break;
        }
        contador=0;
    }//end for
    
    return flag;
}




///Funcao responsavel por compartilhar dados entre populacoes boa com boa ruim com ruim e mediana com mediana
//Parametros Populacao, o rank do processo, e o id do processo para onde vai enviar

void ShareBBMMRR(MPI_Comm communicator, int world_rank, int id, poblacion *pob, permutacion *perm, int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes)
{
    int world_size,i, iii, index,j,prob;
    int flag=0;
    int avaliador=1;
    int * IndicesEnvia;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    IndicesEnvia=buscaIndex(qtdIndividual,typeIndividual,pob->tamanhoPoblacion);// funcao deve ser rescrita pra buscar uma percentual x da populacao
    for (iii=0; iii< qtdIndividual; iii++)
    {
        // If we are the root process, send our data to everyone
        enviar.tamanho=pob->cromosomas[IndicesEnvia[iii]].tamanho;
        enviar.fitness=pob->cromosomas[IndicesEnvia[iii]].fitness;
        
        for (i=0; i<pob->cromosomas[0].tamanho; i++)
            enviar.pi[i]=pob->cromosomas[IndicesEnvia[iii]].genes[i];
        
        /***************Parte onde trata emigracao segundo exclusao ************************/
        
        if (emigracao==2) {
          //  pob->cromosomas[IndicesEnvia[iii]].fitness=pob->cromosomas[0].tamanho; essa eh a linha original
            for(j = 0; j < perm->tamanho; j++){
                prob = rand() % 2;//valores entre 0 y 1
                if (prob == 1)//si prob 1, entonces gen positivo
                    pob->cromosomas[IndicesEnvia[iii]].genes[j] = perm->pi[j];
                else//si prob 0, entonces gen negativo
                    pob->cromosomas[IndicesEnvia[iii]].genes[j] = perm->pi[j] * -1;
            }//end for
            pob->cromosomas[IndicesEnvia[iii]].fitness= calcular_fitness(pob->cromosomas[IndicesEnvia[iii]], &j);// provavel dar erro
            
        }//end if emigracao ==2
        
        
        
        /*******************aqui onde vai ser enviado e recebido os dados********************/
        
        
        
        int error_code= MPI_Send(&enviar,//the data
                                     
                                     sizeof(permutacion),
                                     
                                     MPI_BYTE,
                                     
                                     id,//destination process
                                     
                                     0,//tag
                                     
                                     communicator);
            
            
            
        if (error_code != MPI_SUCCESS)
                
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
      
        error_code=MPI_Recv(&recebe,
                                
                                sizeof(permutacion),
                                
                                MPI_BYTE,
                                
                                 id, // processo que enviou
                                
                                0,//tag
                                
                                communicator,
                                
                                MPI_STATUS_IGNORE);
            
            
            
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
            
        if (flag)
            flag=0;
                
        else
            flag=1;
        index=Imigracao(imigracao, pob->tamanhoPoblacion, recebe.fitness,  pob,  flag);
        
        if (index>=pob->tamanhoPoblacion)
        {
            index=pob->tamanhoPoblacion-1;
            printf("DEU ERRADO NA PARTE DE IMIGRACAO \n");
        }
        avaliador=CompareIndividuals(pob, recebe, Pgenes);
         if (avaliador)
         {
             pob->cromosomas[index].fitness=recebe.fitness;
            for (i=0; i<pob->cromosomas[0].tamanho; i++)
                pob->cromosomas[index].genes[i]=recebe.pi[i];
         }
        
    }//end for

    free(IndicesEnvia);

}// end function ShareBBMMRR


//P0 calcula uma troca aleatoria entre ilhas, uso o modelo de comunicao BBMMRR
void Aleatory2(MPI_Comm communicator, int tamanho,poblacion *pob, permutacion *perm, int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes)
{
    int i=0,x,x1, escravo,error_code,world_rank,alvo=-1;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    int * b= malloc (12 * sizeof (int));
    for (; i<tamanho; i++)
        b[i]=i;
    //vamos trocar posicoes para mesclar a troca de dados entre as ilhas
    
    for (i=0; i<tamanho; i++) {
        x = rand() % 12; //gera aleatorio entre 0 e 11
        x1 = rand() % 12; //gera aleatorio entre 0 e 11
        escravo=b[x];
        b[x]=b[x1];
        b[x1]=escravo;
    }
    /*
    printf("Ordem do arranjo\n");
    for (i=0; i<tamanho; i++)
        printf("%d ", b[i]);
    printf("\n");
    */
    
    for (i=0; i<tamanho; i++)
    {
        if ((i%2==0 && b[i+1]==0) || (i%2==1 && b[i-1]==0) ) {
            i++;
            if (i==tamanho) {
                break;
            }
        }
        
        if (b[i]==0)
        {//recolher dados para enviar compartilhar individuos
            if (i%2==1)
                alvo=b[i-1];
            
            else
                alvo=b[i+1];
            
            
        }//
        if (i%2==0) {
      //      printf("enviei o id %d para o processo %d\n", b[i], b[i+1]);
            int error_code= MPI_Send(&b[i],//the data
                                     
                                     sizeof(int),
                                     
                                     MPI_INT,
                                     
                                     b[i+1],//destination process
                                     
                                     0,//tag
                                     
                                     communicator);
            
            
            
            if (error_code != MPI_SUCCESS)
                
            {
                char error_string[BUFSIZ];
                
                int length_of_error_string;
                
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                printf("%3d: %s\n", world_rank, error_string);
            }//end if
        }//end if
        else{
            //printf("enviei o id %d para o processo %d\n", b[i], b[i-1]);
            error_code= MPI_Send(&b[i],//the data
                                 
                                 sizeof(int),
                                 
                                 MPI_INT,
                                 
                                 b[i-1],//destination process
                                 
                                 0,//tag
                                 
                                 communicator);
            
            
            
            if (error_code != MPI_SUCCESS)
                
            {
                char error_string[BUFSIZ];
                
                int length_of_error_string;
                
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                printf("%3d: %s\n", world_rank, error_string);
            }//end if
            
        }//end else
        
    }//end for
    //printf("meu alvo eh %d\n", alvo);
    //se sou o processo 0 devo esperar por um msg q foi enviada com os individuos q posteriormente serao enviados
    //chamar funcao responsavel pela troca de dados entre o individuos
    if (alvo==-1) {
        printf("*****************Erro o alvo nao foi setado***************\n\n");
    }
    ShareBBMMRR(communicator, world_rank, alvo, pob, perm, emigracao, qtdIndividual, typeIndividual, imigracao, Pgenes);
    free(b);
}




//vamos passar o valor calculado para designar a populacao junto com o processo da mesma alem da propria populacao que vamos usar para o processo 0
// vai retornar uma estrutura com o id do processo e o calculo da media + variancia da respectiva populacao
  void ParaCoordenador2(MPI_Comm communicator, int world_rank, float mediaVariante, poblacion *pob, permutacion *perm, int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes){
    int world_size,error_code,comp;
    MPI_Comm_size(communicator, &world_size);
    if (world_rank!=0)
        
    {
       
        //aqui estou recebendo do processo 0 o id do processo que devo compartilhar meus individuos
        error_code=MPI_Recv(&comp,
                            sizeof(int),
                            MPI_INT,
                            0, // processo que enviou
                            0,//tag
                            communicator,
                            MPI_STATUS_IGNORE);
        
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
        }//end if
        // printf("sou o P %d e recebe %d \n", world_rank, comp);
        
        //chamar funcao responsavel pela troca de dados entre o individuos
        ShareBBMMRR( communicator, world_rank, comp, pob, perm, emigracao, qtdIndividual, typeIndividual, imigracao, Pgenes);
    }//end if
    
    else
    {//sou o processo zero
        
        //vamos trabalhar que os 4 primeiros sao as melhores ilhas as prox 4 sao medianas e 4 ultimas as piores
          Aleatory2(communicator, world_size,    pob, perm, emigracao, qtdIndividual, typeIndividual, imigracao,  Pgenes);
    }//else
}







void GA(MPI_Comm communicator,int world_rank,parameters *params, float prob_crossover,float porcentaje_selec,float porcentaje_reemp,float prob_mutacion, int emigracao, int typeIndividual, int qtdIndividual,int imigracao, float intervaloMigracao, float Pgenes)
{
    permutacion perm;
    poblacion pob;//poblacion
    int mejor_solucion;//mejor solucion encontrada
    int i, num_generaciones,nro_generacion;
    int num_eval_fit_function = 0;
   
    //Read Permutation
    leer_permutacion_de_archivo("i", &perm);
    
    params->size_population = -1;
    params->number_generations = -1;
    params->num_eval_fit_function = -1;
    params->number_reversals = -1;
    
    poblacion_inicial(&pob,&perm,porcentaje_reemp, porcentaje_selec, params);//crear poblacion inicial
    fitness(&pob,1,&num_eval_fit_function);//calcular fitness pob inicial
    mejor_solucion = pob.cromosomas[0].fitness;
    num_generaciones = perm.tamanho;//num de generaciones = a tamanho de permutacion
    int contador = num_generaciones * intervaloMigracao;
    for(nro_generacion=2; nro_generacion<=num_generaciones; nro_generacion++)
    {

        seleccion(&pob,nro_generacion,&mejor_solucion);//seleccion:ordenar poblacion
        crossover3(&pob, &perm, prob_crossover);//crossover:generar descendencia
        mutacion(&pob, &perm, prob_mutacion);//mutacion:actuar sobre toda la descendencia (de tam_desc hasta tamanho de descencia)
        fitness(&pob,nro_generacion,&num_eval_fit_function);//calcular fitness de descendencia
        reemplazo(&pob);//reemplazar descendencia en nueva generacion
        
        contador--;

        if (contador==0)
        {
            ParaCoordenador2(communicator,world_rank, avaliaPopulation(&pob),&pob, &perm,  emigracao,  qtdIndividual,  typeIndividual,  imigracao,  Pgenes);
            MPI_Barrier(MPI_COMM_WORLD);
            contador = num_generaciones * intervaloMigracao;
        }
        
        if (nro_generacion==num_generaciones)
        {
            sendALL(communicator, world_rank, &pob);
            
        }//end if

       num_eval_fit_function=0;
    }//end for
    
    
    if (world_rank==0)
    {    seleccion(&pob,nro_generacion,&mejor_solucion);//seleccion:ordenar poblacion
        printf("RandomNP %d\n",mejor_solucion);
    }
    
    for(i=0;i<pob.tamanhoTotal;i++){
        free(pob.cromosomas[i].genes);
    }
    free(pob.cromosomas);

    
}



int main(int argc,char **argv)
{
    int seed[24] = {408302,680502,414942,829352,627420, 753780, 110291,444847, 127844, 525817,887214,115510,112376,510836,310830,390833,560813,260810,908811,414819,939819,287817,364410,194419};
    unsigned int seed2 = time(NULL);
    float prob_crossover=      0.97;
    float porcentaje_selec=    0.96;	
    float porcentaje_reemp= 	0.62;
    float prob_mutation= 	0.01;
    int emigracao= 		2;
    int typeIndividual= 	1;
    int qtdIndividual= 		6;
    int imigracao=		1;
    float intervaloMigracao= 	0.88;
    float PPgenes=		0.84;

    int world_rank;
    MPI_Init(NULL, NULL);
    //printf("quantos individuos vou enviar %d\n",qtdIndividual);
    parameters params;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    srand (seed[world_rank]+seed2);//initialize random generator
    params.seed =seed[world_rank]+seed2;
    GA(MPI_COMM_WORLD, world_rank, &params, prob_crossover,porcentaje_selec,porcentaje_reemp, prob_mutation,emigracao,typeIndividual,qtdIndividual,imigracao,intervaloMigracao, PPgenes);
    MPI_Finalize();
    return 0;
}






