/*
 ============================================================================
 Project    : Parallel Genetic Algorithm for the SUPBR Problem
                (version with multiple populations)
 File       : main.c
 Author     : Jose Luis Soncco-Alvarez and Mauricio Ayala-Rincon
                Group of Theory of Computation
                Universidade de Brasilia (UnB) - Brazil
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>//time
#include <mpi.h>
#include <math.h>
#include <string.h>
#include "Tree.c"
#include "operadores.h"
#include "structs_ga.h"
#include "calc_fitness.h"
#include "GerarPopuInicial.c"


#define WORKTAG 1
#define DIETAG 2
#define EVALTAG 3
#define GOTAG 4

#define DEBUG 0
#define TAM_BUFFER 1000

//void leer_permutacion(permutacion *perm);
void slave(parameters *params);
/*
void leer_permutacion(permutacion *perm){
	int i,n;
	scanf("%d",&n);//leer tamanho de permutacion
	perm->tamanho = n;
	for(i=0; i<n; i++){
		scanf("%d",&perm->pi[i]);
	}
}
*/

//flag em uma roda vai comecar no inicio outro no final
int Imigracao( int imigracao, int sizePopulation, int fitSemelhante, poblacion *pob, int flag){
    int i,indice=0;
    int porcento=sizePopulation *0.2;
    switch (imigracao)
    {
        case 1: // pior
            indice= (sizePopulation - porcento) +(rand() % porcento);
            break;
        case 2: //aleatorio
            indice=(rand() % sizePopulation);
            break;
        case 3://semelhante
            indice=(rand() % sizePopulation);
            if (flag)
            {
                for (i=0; i<sizePopulation; i++)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end if
            else
            {
                for (i=(sizePopulation-1); i>=0; i--)
                {
                    if (pob->cromosomas[i].fitness==fitSemelhante)
                    {
                        indice=i;
                        break;
                    }
                }//end for
            }//end else
            break;
    }//end switch
    
    return indice;
    
    
}// end function Imigracao


//busca os indicies a ser enviado para outra ilha
int * buscaIndex(int qtdIndividual, int typeIndividual, int sizePopulation)
{
    int i,j;
    int * indices= malloc (qtdIndividual * sizeof (int));
    if (indices==NULL)
        printf("ATENCAO PERCA DE MEMORIA, GA_Paralle_pop.c, LINE 236\n");
    switch (typeIndividual) {
        case 1://melhor
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=i;
            }
            break;
            
        case 2: // pior
            i=0;
            j=(sizePopulation-1);
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=j;
                j--;
                
            }//end for
            break;
            
        case 3: //aleatorio
            for (i=0; i<qtdIndividual; i++)
            {
                indices[i]=(rand() % sizePopulation);
            }
            break;
    }//end switch
    
    
    return indices;
    
}// end buscaIndex




void sendALL( MPI_Comm communicator, int world_rank, poblacion *pob)
{
    
    int world_size, j, iii,i;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    
    // If we are the root process, send our best data to everyone
    enviar.tamanho=pob->cromosomas[0].tamanho;
    enviar.fitness=pob->cromosomas[0].fitness;
    for (i=0; i<pob->cromosomas[0].tamanho; i++)
        enviar.pi[i]=pob->cromosomas[0].genes[i];
    
    
    
    if (world_rank!=0)
        
    {
        int error_code= MPI_Send(&enviar,//the data
                                 sizeof(permutacion),
                                 MPI_BYTE,
                                 0,//destination process
                                 0,//tag
                                 communicator);
        if (error_code != MPI_SUCCESS)
        {
            char error_string[BUFSIZ];
            int length_of_error_string;
            MPI_Error_string(error_code, error_string, &length_of_error_string);
            printf("%3d: %s\n", world_rank, error_string);
            
        }
        
    }
    
    else
    {
        for( iii = 0; iii < world_size; iii++)
        {
            
            
            if (iii != world_rank)
            {
                int error_code=MPI_Recv(&recebe,
                                        sizeof(permutacion),
                                        MPI_BYTE,
                                        iii, // processo que enviou
                                        0,//tag
                                        communicator,
                                        MPI_STATUS_IGNORE);
                
                if (error_code != MPI_SUCCESS)
                {
                    char error_string[BUFSIZ];
                    int length_of_error_string;
                    MPI_Error_string(error_code, error_string, &length_of_error_string);
                    printf("%3d: %s\n", world_rank, error_string);
                }//end if
                
                int controle=0;
                for (j=0; j<pob->tamanhoPoblacion; j++)
                {
                    if (recebe.fitness < pob->cromosomas[j].fitness)
                    {
                        pob->cromosomas[j].fitness=recebe.fitness;
                        for (i=0; i<pob->cromosomas[j].tamanho; i++)
                        {
                            pob->cromosomas[j].genes[i]=recebe.pi[i];
                        }//end for
                        controle=1;
                    }//end if
                    if (controle)
                        break;
                }//end for
            }//end if
        }//end for
        
    }//else
    
}// end function





void mostrar_fitness(poblacion *pob, int tamanho){
    
    int i;
    //imprimir poblacion actual
    for(i=pob->basePoblacion; i<pob->topePoblacion; i++){
        printf("(%d): ",i);
        
            printf("%d) %d ", i, pob->cromosomas[i].fitness);
        printf("\n");
    }
    
}

//return 1 to insert and 0 to discard
// Pgene  vai de 0.1 ate 1
// uma flag sera imposta a fim de configura a quantidade da populacao a ser comparada
int CompareIndividuals(poblacion *pob, permutacion genomaAlvo, float Pgenes) {
    int i,j;
    int flag=1;
    int contador=0;
    int intervalo=pob->tamanhoPoblacion * Pgenes;
    //printf("intervalo levado em consideracao %d sobre %d\n ", intervalo, pob->sizePopulation);
    for (i=0; i<intervalo; i++) {
        
        for (j=0; j<pob->cromosomas[i].tamanho; j++)
        {
            if(pob->cromosomas[i].genes[j]==genomaAlvo.pi[j])
            {    contador++;
            }
        }//end for
        
        if (contador==pob->cromosomas[i].tamanho) {
            flag=0;
            break;
        }
        contador=0;
    }//end for
    
    return flag;
}
//implementa topologia tree
//flag é pra marcar se for igual a 1 significa q esta terminando as geracoes e todos as ilhas necessitas saber os resultados
void Topologia24nodos( MPI_Comm communicator, int world_rank, poblacion *pob,int emigracao, int qtdIndividual, int typeIndividual, int imigracao, float Pgenes){
     int world_size,ii,i, iii, index;
    int flag=0;
    
    int avaliador=1;
    int * IndicesEnvia;
    MPI_Comm_size(communicator, &world_size);
    permutacion enviar,recebe;
    IndicesEnvia=buscaIndex(qtdIndividual,typeIndividual,pob->tamanhoPoblacion);
    /*************Setando pra quais nodos vai ser enviado os dados e recebidos******************/
    int * msg= tree24(world_rank);
    for (iii=0; iii< qtdIndividual; iii++)
    {
        // If we are the root process, send our data to everyone
        enviar.tamanho=pob->cromosomas[IndicesEnvia[iii]].tamanho;
        enviar.fitness=pob->cromosomas[IndicesEnvia[iii]].fitness;
        
        for (i=0; i<pob->cromosomas[0].tamanho; i++)
            enviar.pi[i]=pob->cromosomas[IndicesEnvia[iii]].genes[i];
        
        /***************Parte onde trata emigracao segundo exclusao ************************/
        
        if (emigracao==2) {
            pob->cromosomas[IndicesEnvia[iii]].fitness=pob->cromosomas[0].tamanho;
        }
        
        
        
        /*******************aqui onde vai ser enviado e recebido os dados********************/
        
        
        for (ii = 0; ii < 4; ii++)
            
        {
            if(msg[ii]==-1)
                break;
            
            int error_code= MPI_Send(&enviar,//the data
                                     
                                     sizeof(permutacion),
                                     
                                     MPI_BYTE,
                                     
                                     msg[ii],//destination process
                                     
                                     0,//tag
                                     
                                     communicator);
            
            
            
            if (error_code != MPI_SUCCESS)
                
            {
                
                char error_string[BUFSIZ];
                
                int length_of_error_string;
                
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                printf("%3d: %s\n", world_rank, error_string);
                
                
                
            }//end if
            
            
            
            error_code=MPI_Recv(&recebe,
                                
                                sizeof(permutacion),
                                
                                MPI_BYTE,
                                
                                msg[ii], // processo que enviou
                                
                                0,//tag
                                
                                communicator,
                                
                                MPI_STATUS_IGNORE);
            
            
            
            if (error_code != MPI_SUCCESS)
                
            {
                
                char error_string[BUFSIZ];
                
                int length_of_error_string;
                
                MPI_Error_string(error_code, error_string, &length_of_error_string);
                
                printf("%3d: %s\n", world_rank, error_string);
                
            }//end if
            
            if (flag) {
                
                flag=0;
                
            }
            
            else
                
                flag=1;
            
            index=Imigracao(imigracao, pob->tamanhoPoblacion, recebe.fitness,  pob,  flag);
            
            // printf("index %d tamanho total %d\n", index,pob->sizePopulation);
            
            if (index>=pob->tamanhoPoblacion) {
                
                index=pob->tamanhoPoblacion-1;
                
                printf("DEU ERRADO NA PARTE DE IMIGRACAO \n");
                
            }
            
            avaliador=CompareIndividuals(pob, recebe, Pgenes);
            
            if (avaliador) {
                
                pob->cromosomas[index].fitness=recebe.fitness;
                
                for (i=0; i<pob->cromosomas[0].tamanho; i++)
                    pob->cromosomas[index].genes[i]=recebe.pi[i];
            }
        }//end for
        
    }//end for
    
    
    free(msg);
    free(IndicesEnvia);
    
    
}//end function tree 24


void GA(MPI_Comm communicator,int world_rank,parameters *params, float prob_crossover,float porcentaje_selec,float porcentaje_reemp,float prob_mutacion, int emigracao, int typeIndividual, int qtdIndividual,int imigracao, float intervaloMigracao, float Pgenes)
{
    permutacion perm;
    poblacion pob;
    int mejor_solucion=-1;//mejor solucion encontrada
    int i, num_generaciones,nro_generacion, world_size;
    MPI_Comm_size(communicator, &world_size);
    
    //Read Permutation
    leer_permutacion_de_archivo("i", &perm);
    params->size_population = -1;
    params->number_generations = -1;
    params->num_eval_fit_function = -1;
    params->number_reversals = -1;
    
    //poblacion_inicial(&pob,&perm,porcentaje_reemp, porcentaje_selec, params);//crear poblacion inicial
   //construir a populacao
    construirPopInit(world_size,world_rank);
    //carregar populacao
    CarregarPopInitial(world_rank, porcentaje_reemp, porcentaje_selec, &pob,&perm);
    
    int num_eval_fit_function = 0;
    fitness(&pob,1,&num_eval_fit_function);//calcular fitness pob inicial
    //mostrar_fitness(&pob, perm.tamanho);
    mejor_solucion = pob.cromosomas[0].fitness;
    
  
    num_generaciones = perm.tamanho;//num de generaciones = a tamanho de permutacion
    int contador = num_generaciones * intervaloMigracao;
    for(nro_generacion=2; nro_generacion<=num_generaciones; nro_generacion++)
    {
        
        seleccion(&pob,nro_generacion,&mejor_solucion);//seleccion:ordenar poblacion
        crossover3(&pob, &perm, prob_crossover);//crossover:generar descendencia
        mutacion(&pob, &perm, prob_mutacion);//mutacion:actuar sobre toda la descendencia (de tam_desc hasta tamanho de descencia)
        fitness(&pob,nro_generacion,&num_eval_fit_function);//calcular fitness de descendencia
        reemplazo(&pob);//reemplazar descendencia en nueva generacion
        
        contador--;
        if (contador==0)
        {
           Topologia24nodos( communicator, world_rank, &pob, emigracao,qtdIndividual, typeIndividual,imigracao, Pgenes);
            
            contador = num_generaciones * intervaloMigracao;
        }
        
        if (nro_generacion==num_generaciones)
        {
            sendALL(communicator, world_rank, &pob);
            
        }//end if
        
        MPI_Barrier(MPI_COMM_WORLD);
        num_eval_fit_function=0;
    }
    
   
    if (world_rank==0)
    {    seleccion(&pob,nro_generacion,&mejor_solucion);//seleccion:ordenar poblacion
        printf("TreeP %d\n",mejor_solucion);
    }
   
    for(i=0;i<pob.tamanhoTotal;i++){
        free(pob.cromosomas[i].genes);
    }
    free(pob.cromosomas);

    
}

int main(int argc,char **argv)
{
    int seed[24] = {408302,680502,414942,829352,627420, 753780, 110291,444847, 127844, 525817,887214,115510,112376,510836,310830,390833,560813,260810,908811,414819,939819,287817,364410,194419};
    unsigned int seed2 = time(NULL);
    float prob_crossover=	0.94;
    float porcentaje_selec=	0.94;
    float porcentaje_reemp=	0.5;
    float prob_mutation=	0.01;
    int emigracao=		1;
    int typeIndividual=		3;
    int qtdIndividual=		2;
    int imigracao=		2;
    float intervaloMigracao=	0.1;
    float Pgenes           =    0.5;
    int world_rank;
    MPI_Init(NULL, NULL);
    parameters params;
    MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
    srand (seed2);//initialize random generator
    params.seed =seed[world_rank]+seed2;
    GA(MPI_COMM_WORLD, world_rank, &params, prob_crossover,porcentaje_selec,porcentaje_reemp, prob_mutation,emigracao,typeIndividual,qtdIndividual,imigracao,intervaloMigracao,Pgenes);
    MPI_Finalize();
    return 0;
}






